@@ -1,5 +1,7 @@
 {-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE PatternSynonyms #-}
+{-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE MultiWayIf #-}
 
 import Control.Applicative
 import Control.Monad
@@ -42,24 +44,56 @@ unsafeModules = map modToFile
   , "Codata.Musical.Covec"
   , "Codata.Musical.Conversion"
   , "Codata.Musical.Stream"
+  , "Data.Bytestring.Base"
+  , "Data.Bytestring.Builder.Base"
+  , "Data.Bytestring.Builder.Primitive"
+  , "Data.Bytestring.IO"
+  , "Data.Bytestring.IO.Primitive"
+  , "Data.Bytestring.Primitive"
+  , "Data.Word8.Base"
+  , "Data.Word8.Literals"
+  , "Data.Word8.Primitive"
+  , "Data.Word64.Primitive"
+  , "Data.Word8.Show"
+  , "Data.Word64.Show"
   , "Debug.Trace"
+  , "Effect.Monad.IO"
+  , "Effect.Monad.IO.Instances"
   , "Foreign.Haskell"
   , "Foreign.Haskell.Coerce"
   , "Foreign.Haskell.Either"
   , "Foreign.Haskell.Maybe"
+  , "Foreign.Haskell.List.NonEmpty"
   , "Foreign.Haskell.Pair"
   , "IO"
   , "IO.Base"
+  , "IO.Categorical"
+  , "IO.Handle"
   , "IO.Infinite"
+  , "IO.Instances"
+  , "IO.Effectful"
   , "IO.Finite"
   , "IO.Primitive"
+  , "IO.Primitive.Core"
+  , "IO.Primitive.Handle"
   , "IO.Primitive.Infinite"
   , "IO.Primitive.Finite"
   , "Relation.Binary.PropositionalEquality.TrustMe"
+  , "System.Clock"
+  , "System.Clock.Primitive"
+  , "System.Directory"
+  , "System.Directory.Primitive"
   , "System.Environment"
   , "System.Environment.Primitive"
   , "System.Exit"
   , "System.Exit.Primitive"
+  , "System.FilePath.Posix"
+  , "System.FilePath.Posix.Primitive"
+  , "System.Process"
+  , "System.Process.Primitive"
+  , "System.Random"
+  , "System.Random.Primitive"
+  , "Test.Golden"
   , "Text.Pretty.Core"
   , "Text.Pretty"
   ] ++ sizedTypesModules
@@ -81,8 +115,8 @@ withKModules = map modToFile
   , "Data.Star.Pointer"
   , "Data.Star.Vec"
   , "Data.String.Unsafe"
-  , "Reflection.Annotated"
-  , "Reflection.Annotated.Free"
+  , "Reflection.AnnotatedAST"
+  , "Reflection.AnnotatedAST.Free"
   , "Relation.Binary.HeterogeneousEquality"
   , "Relation.Binary.HeterogeneousEquality.Core"
   , "Relation.Binary.HeterogeneousEquality.Quotients.Examples"
@@ -103,36 +137,40 @@ isWithKModule =
 
 sizedTypesModules :: [FilePath]
 sizedTypesModules = map modToFile
-  [ "Codata.Cofin"
-  , "Codata.Cofin.Literals"
-  , "Codata.Colist"
-  , "Codata.Colist.Bisimilarity"
-  , "Codata.Colist.Categorical"
-  , "Codata.Colist.Properties"
-  , "Codata.Conat"
-  , "Codata.Conat.Bisimilarity"
-  , "Codata.Conat.Literals"
-  , "Codata.Conat.Properties"
-  , "Codata.Covec"
-  , "Codata.Covec.Bisimilarity"
-  , "Codata.Covec.Categorical"
-  , "Codata.Covec.Instances"
-  , "Codata.Covec.Properties"
-  , "Codata.Cowriter"
-  , "Codata.Cowriter.Bisimilarity"
-  , "Codata.Delay"
-  , "Codata.Delay.Bisimilarity"
-  , "Codata.Delay.Categorical"
-  , "Codata.Delay.Properties"
-  , "Codata.M"
-  , "Codata.M.Bisimilarity"
-  , "Codata.M.Properties"
-  , "Codata.Stream"
-  , "Codata.Stream.Bisimilarity"
-  , "Codata.Stream.Categorical"
-  , "Codata.Stream.Instances"
-  , "Codata.Stream.Properties"
-  , "Codata.Thunk"
+  [ "Codata.Sized.Cofin"
+  , "Codata.Sized.Cofin.Literals"
+  , "Codata.Sized.Colist"
+  , "Codata.Sized.Colist.Bisimilarity"
+  , "Codata.Sized.Colist.Categorical"
+  , "Codata.Sized.Colist.Effectful"
+  , "Codata.Sized.Colist.Properties"
+  , "Codata.Sized.Conat"
+  , "Codata.Sized.Conat.Bisimilarity"
+  , "Codata.Sized.Conat.Literals"
+  , "Codata.Sized.Conat.Properties"
+  , "Codata.Sized.Covec"
+  , "Codata.Sized.Covec.Bisimilarity"
+  , "Codata.Sized.Covec.Categorical"
+  , "Codata.Sized.Covec.Effectful"
+  , "Codata.Sized.Covec.Instances"
+  , "Codata.Sized.Covec.Properties"
+  , "Codata.Sized.Cowriter"
+  , "Codata.Sized.Cowriter.Bisimilarity"
+  , "Codata.Sized.Delay"
+  , "Codata.Sized.Delay.Bisimilarity"
+  , "Codata.Sized.Delay.Categorical"
+  , "Codata.Sized.Delay.Effectful"
+  , "Codata.Sized.Delay.Properties"
+  , "Codata.Sized.M"
+  , "Codata.Sized.M.Bisimilarity"
+  , "Codata.Sized.M.Properties"
+  , "Codata.Sized.Stream"
+  , "Codata.Sized.Stream.Bisimilarity"
+  , "Codata.Sized.Stream.Categorical"
+  , "Codata.Sized.Stream.Effectful"
+  , "Codata.Sized.Stream.Instances"
+  , "Codata.Sized.Stream.Properties"
+  , "Codata.Sized.Thunk"
   , "Data.Container.Fixpoints.Sized"
   , "Data.W.Sized"
   , "Data.Nat.PseudoRandom.LCG.Unsafe"
@@ -207,10 +245,10 @@ extractHeader mod = extract
 
 -- | A crude classifier looking for lines containing options
 
-data Status = Deprecated | Unsafe | Safe
-  deriving (Eq)
+data Safety = Unsafe | Safe deriving (Eq)
+data Status = Deprecated | Active deriving (Eq)
 
-classify :: FilePath -> [String] -> [String] -> Exc Status
+classify :: FilePath -> [String] -> [String] -> Exc (Safety, Status)
 classify fp hd ls
   -- We start with sanity checks
   | isUnsafe && safe          = throwError $ fp ++ contradiction "unsafe" "safe"
@@ -219,11 +257,12 @@ classify fp hd ls
   | isWithK && not withK      = throwError $ fp ++ missingWithK
   | not (isWithK || cubicalC) = throwError $ fp ++ uncategorized "as relying on K" "cubical-compatible"
   -- And then perform the actual classification
-  | deprecated                = pure $ Deprecated
-  | isUnsafe                  = pure $ Unsafe
-  | safe                      = pure $ Safe
-  -- We know that @not (isUnsafe || safe)@, all cases are covered
-  | otherwise                 = error "IMPOSSIBLE"
+  | otherwise = do
+      let safety = if | safe -> Safe
+                      | isUnsafe -> Unsafe
+                      | otherwise -> error "IMPOSSIBLE"
+      let status = if deprecated then Deprecated else Active
+      pure (safety, status)
 
   where
 
@@ -260,18 +299,20 @@ classify fp hd ls
 data LibraryFile = LibraryFile
   { filepath   :: FilePath -- ^ FilePath of the source file
   , header     :: [String] -- ^ All lines in the headers are already prefixed with \"-- \".
-  , status     :: Status   -- ^ Safety options used by the module
+  , safety     :: Safety
+  , status     :: Status   -- ^ Deprecation status options used by the module
   }
 
 analyse :: FilePath -> IO LibraryFile
 analyse fp = do
   ls <- lines <$> readFileUTF8 fp
   hd <- runExc $ extractHeader fp ls
-  cl <- runExc $ classify fp hd ls
+  (sf, st) <- runExc $ classify fp hd ls
   return $ LibraryFile
-    { filepath   = fp
-    , header     = hd
-    , status     = cl
+    { filepath = fp
+    , header   = hd
+    , safety   = sf
+    , status   = st
     }
 
 checkFilePaths :: String -> [FilePath] -> IO ()
@@ -280,17 +321,37 @@ checkFilePaths cat fps = forM_ fps $ \ fp -> do
   unless b $
     die $ fp ++ " is listed as " ++ cat ++ " but does not exist."
 
+data Options = Options
+  { includeDeprecated :: Bool
+  , outputDirectory :: FilePath
+  }
+
+initOptions :: Options
+initOptions = Options
+  { includeDeprecated = False
+  , outputDirectory = "."
+  }
+
+parseOptions :: [String] -> Options -> Maybe Options
+parseOptions [] opts = pure opts
+parseOptions ("--include-deprecated" : rest) opts
+  = parseOptions rest (opts { includeDeprecated = True })
+parseOptions ("--out-dir" : dir : rest) opts
+  = parseOptions rest (opts { outputDirectory = dir })
+parseOptions _ _ = Nothing
+
 ---------------------------------------------------------------------------
 -- Collecting all non-Core library files, analysing them and generating
--- 4 files:
+-- 2 files:
 -- Everything.agda                 all the modules
 -- EverythingSafe.agda             all the safe modules
 
+main :: IO ()
 main = do
   args <- getArgs
-  case args of
-    [] -> return ()
-    _  -> hPutStr stderr usage >> exitFailure
+  Options{..} <- case parseOptions args initOptions of
+    Just opts -> pure opts
+    Nothing -> hPutStr stderr usage >> exitFailure
 
   checkFilePaths "unsafe" unsafeModules
   checkFilePaths "using K" withKModules
@@ -300,22 +361,23 @@ main = do
                find always
                     (extension ==? ".agda" ||? extension ==? ".lagda")
                     srcDir
-  libraryfiles <- filter ((Deprecated /=) . status) <$> mapM analyse modules
+  libraryfiles <- (if includeDeprecated then id
+    else (filter ((Deprecated /=) . status) <$>)) (mapM analyse modules)
 
   let mkModule str = "module " ++ str ++ " where"
 
-  writeFileUTF8 (allOutputFile ++ ".agda") $
+  writeFileUTF8 (outputDirectory ++ "/" ++ allOutputFile ++ ".agda") $
     unlines [ header
             , "{-# OPTIONS --rewriting --guardedness --sized-types #-}\n"
             , mkModule allOutputFile
             , format libraryfiles
             ]
 
-  writeFileUTF8 (safeOutputFile ++ ".agda") $
+  writeFileUTF8 (outputDirectory ++ "/" ++ safeOutputFile ++ ".agda") $
     unlines [ header
             , "{-# OPTIONS --safe --guardedness #-}\n"
             , mkModule safeOutputFile
-            , format $ filter ((Unsafe /=) . status) libraryfiles
+            , format $ filter ((Unsafe /=) . safety) libraryfiles
             ]
 
 -- | Usage info.
@@ -332,6 +394,9 @@ usage = unlines
   , "The program generates documentation for the library by extracting"
   , "headers from library modules. The output is written to " ++ allOutputFile
   , "with the file " ++ headerFile ++ " inserted verbatim at the beginning."
+  , ""
+  , "If the option --out-dir is used then the output is placed in the"
+  , "subdirectory thus selected."
   ]
 
 
